\ TED.F
\
\ TED - A Tiny Text Editor for DX-Forth
\
\ Based on the HT-68K editor by J.Bartel
\
\ The HELP screen is only compiled for the
\ turnkey version
\
\    ^E   Up cursor      ^R   Prev page
\    ^X   Down cursor    ^C   Next page
\    ^D   Right cursor   ^G   Del next char
\    ^S   Left cursor    ^H   Del prev char
\    ^QS  Home cursor  	 ^QD  EOL cursor
\    ^QR  Begin cursor   ^QC  EOF cursor
\    ^M   New line       ^T   Erase to EOL
\    ^Y   Delete line    ^L   Restore line
\    ^I   Indent         ^J   Help
\    ^Z   Clear text     ^V   Ins/Ovr
\    ^B   Redraw screen  ^N   Line number  
\    ^KS  Save 
\    ^KQ or ESC  Quit    ^KX  Exit with save
\
\ Revision
\ 2015-06-03 es  Specify a filename
\ 2015-12-09 es  Join lines with ^G or DEL
\ 2016-09-25 es  Updated for DX-Forth
\ 2017-01-25 es  Replace EXIT THEN with END
\ 2017-08-16 es  Add ^Q & updated for DX-Forth
\ 2018-01-17 es  Updated for DX-Forth
\ 2019-11-19 es  Minor bug fixes
\ 2020-02-26 pz  - fix ^T
\                - add ^J, ^V, ^N, ^B, esc,
\                  ^QS/D/R/C, ^KQ/X,
\                - change ^I, ^Z, key remapping for DOS
\                - add: start cursor position is set to
\                  leaving position at previous session

forth definitions  decimal

false \ true for turnkey

( *) dup [if] application [then]

cr .( loading TED Text Editor )

\ Running DX-Forth for CP/M or DOS?
$111 @ $4683 = constant CPM?

\ convert null terminated string (adr1) to string (adr2 u)
: ZCOUNT \ adr1 -- adr2 u ;
  dup -1 0 scan drop over -
;

: ZPLACE \ adr1 u adr2 -- ; move string (adr1 u) to adr2 as null terminated
  2dup + >r swap cmove 0 r> c!
;

\ Video terminal specific
79 value XMAX  \ #columns - 1
24 value YMAX  \ #rows - 1

\ INSERT-LINE ( -- )  insert blank line at cursor;
\                     remaining rows scroll down
\ DELETE-LINE ( -- )  delete line at cursor;
\                     remaining rows scroll up
\ CLEAR-LINE ( -- )   blank from cursor to end of line

\ pointer operations
: 1+! ( a -- )     1 swap +! ;
: 1-! ( a -- )    -1 swap +! ;
: C@+ ( a -- c )  dup @ c@ swap 1+! ;
: C!+ ( c a -- )  tuck @ c! 1+! ;
: -C@ ( a -- c )  dup 1-! @ c@ ;
: -C! ( c a -- )  dup 1-! @ c! ;

\ max line length
132 constant COLS
\ max printable char mask ($FF for 32-255, $7F for 32-127)
$7F constant MAXM
\ TAB mask: (power of 2) - 1; usually 1, 3, 7
1 constant TABM

0 value YBOT   \ edit bottom row
0 value BUF    \ edit buffer addr
0 value BUFE   \ edit buffer end + 1
0 value TBUF   \ text buffer addr
0 value LINES  \ line count
0 value MEM    \ top of memory
0 value FNAM   \ filename buffer addr
0 value FID    \ file handle
0 value OVR    \ overwrite/insert

variable COL   \ current column#
variable LIN   \ current line#
variable LADR  \ current line addr
variable LTOP  \ absolute line# at top of screen
variable LPOS  \ current line# relative to top of screen
variable BPOS  \ address of char in edit buffer
variable NXT   \ next free addr in text (contains 0)
variable UPD   \ edit buffer change flag
variable XF    \ exit flag
variable TWC   \ text was changed flag
cpm? not [if]
variable CHAR2 \ second char for key remapping
[then]

: >CTRL \ c -- ctrl-c ;
  $1F and
;

: UKEY \ -- c ;
  key upcase
;

: LMAX \ -- n ;
  lines 1- 0 max
;

: GOXY \ x y -- ;
  1+ at-xy
;

: CURH \ -- ;
  0 0 at-xy
;

: CXY  \ -- ;
  col @ XMAX min   lpos @   goxy
;

: CKEY \ n -- c ; n = char for print
  curh [char] ^ emit emit
  key >ctrl
  curh space space cxy
;

: MSG  \ -- ;
  curh clear-line
;

: CHGD \ -- ;
  upd on
;

: BUFL \ adr -- adr u ; buffer remainder as string
  bufe over -
;

system -?
: [ctrl] \ -- c ;
  char >ctrl [compile] literal
; immediate
application

: AKEY \ -- ;
  ." Press a key to continue "   key drop
;

: .FIL \ -- ; print filename
\  fnam count   20 over min >r + r@ - r> type
  fnam count   dup 20 - 0 max /string    type
;

: .POS \ -- ; print row and column
  13 0 at-xy   lin @ 1+ u.
  22 0 at-xy   col @ 1+ u.
  cxy
;
  
: .OVR \ -- ; print overwrite/insert mode
  27 0 at-xy
  ovr if ." Ovr" else ." Ins" then
  cxy
;

: .HD \ -- ; print header line
  msg   10 0 at-xy
  ." Ln       Cl            File "
  .fil   .pos   .ovr
;

: .ERR \ a u -- ; error message
  msg   .fil   space space  type  akey  .hd
;

: SURE? \ a u -- f ;
  msg  type  ." Are you sure? " y/n
;

: LINE \ -- a u ; actual line as string
  ladr @ zcount
;

: .LINE \ -- ; print actual line
  line XMAX 1+ min type
;

: .RT \ -- ; display string right of cursor
  bpos @  bufl  XMAX 1+  bpos @  buf - - min
  0 max type
;

: ROOM? \ -- f ; is room in buffer?
  bufe 1- c@ bl =
;

: LINE@ \ -- lin adr ; fetch line data
  lin @   ladr @
;

: LINE! \ lin adr -- ; store line data
  ladr !   lin !
;

: GOTOP \ -- ;
  tbuf ladr !   lin off
;

: CURTOP \ -- ;
  gotop   ltop off   col off   lpos off
;
  
: CTRC \ -- ; clear text, reset cursor
  tbuf  dup 1- 3 erase ( nulls )
  1+ nxt !   1 to lines  curtop
;

: -TXT \ -- ; clear text, filename, reset cursor
  ctrc   0 fnam c!
;

: SETUP \ -- ;
[ cpm? ]  [if]
  $168 c@ 1-  to XMAX
  $169 c@ 1-  dup to YMAX  2- to YBOT
[else]
  get-window \ x1 y1 x2 y2 ;
  rot -  dup to YMAX  2- to YBOT
  swap -  to XMAX
[then]
  application  here unused + to mem  pad 80 +
  dup to fnam  80 +  dup to buf  COLS +  dup to bufe
  2+  dup to tbuf  mem u> abort" no space"  -txt
  upd off   twc off   false to ovr
;

: INSC \ c -- ; insert char to buffer
  bpos @  dup 1+  bufl 1+ cmove>
  bpos c!+
;

: LU \ -- ; go up one line in text
  lin 1-!
  ladr  dup 1-!  begin  dup -c@  0= until  1+!
;

: LD \ -- ; go down one line in text
  lin 1+!  ladr begin  dup c@+  0= until drop
;

: SETLIN \ n -- ; setup for line n
  tbuf  over 0 ?do  zcount + 1+  loop
  ladr !  lin !
;

: LINES+ \ -- ;
  lines 1+ to lines
;

: LINES- \ -- ;
  lines 1- to lines
;

: MEM? \ -- ; test available memory
  nxt @  mem u< not if  s" no space" .err  then
;

: REPL \ a u -- ; replace line in text
  >r  line  r@ over - >r
  over +  dup r@ +  nxt @ 1+  dup >r
  2 pick - move  2r> +  dup off  nxt !
  r> cmove  MEM?
;

: BSTR \ -- a u ; buffer as string
  buf bufl -trailing
;

: LEAV \ -- ; leave actual line
  upd @ if  bstr repl   twc on  then  upd off
;

: ENTER \ -- ; start changes on this line
  line  buf  dup COLS blank  swap cmove
  buf  col @ +  bpos !  upd off
;

: .ALL \ -- ; update all screen
  leav  enter
  page  line@
  ltop @  dup setlin
  lmax swap -  YBOT min
  1+ 0 ?do  0 i goxy  .line  ld  loop
  line!  .hd  cxy
;

: SLN \ ltop lin -- ;
  >r  0 max  lmax min  dup  r> max  lmax min
  dup setlin  over -  lpos !  ltop !  .all
;
  
: GLN \ n -- ; go to line#
  dup 7 - swap sln
;

: PU \ -- ; ^R page up
  lin @ if
    leav  ltop @  YBOT -  lin @  YBOT -  sln
  then
;

: PD \ -- ; ^C page dn
  lin @  lines < if
    leav  ltop @  YBOT +  lin @  YBOT +  sln
  then
;
  
: BOF \ -- ; ^QR begin of file
  leav   col off   0 0 sln
;

: EOF \ -- ; ^QC end of file
  leav   col off   lines YBOT - lines sln
;

: LNUM \ -- ; ^N line number
  msg ." Line number: "
  pad dup 5 accept   /number 2drop D>S 1- \ enter line number
  leav   col off   gln
;

: SU \ -- ; scroll up, new line at bottom
  0 0 goxy  delete-line  0 YBOT  dup lpos !  goxy
;

: SD \ -- ; scroll down, new line at top
  0 0 goxy  insert-line  lpos off
;

: UPC \ -- ; ^E line up
  lin @ if
    leav  lu
    lin @  ltop @  1-  = if
      sd  .line  ltop 1-!
    else
      lpos 1-!
    then
    enter  .pos
  then
;

: DN \ -- ; ^X line dn
  lin @  lines < if
    leav  ld
    lin @  ltop @  YBOT 1+  +  = if
      su  .line  ltop 1+!
    else
      lpos 1+!
    then
    enter  .pos
  then
;

: RT \ -- ; ^D right
  col @  XMAX <  if
    col 1+!  bpos 1+!  .pos
  then
;

: LFT \ -- ; ^S left
  col @  if
    bpos 1-!  col 1-!  .pos
  then
;

: HOM \ -- ; ^QS home
  buf bpos !   col off   .pos
;

: EOL \ -- ; ^QD end of line
  bstr dup col !   + bpos !   .pos
;

: NLN \ -- ; ^M new line
  room? if
    13 insc  chgd  leav
    ladr  begin  dup c@+  13 = until  0 over -c!  1+!
    lines+  lin 1+!
    clear-line  col off  enter
    lpos @  YBOT  = if
      su  ltop 1+!
    else
      lpos 1+!  insert-line  cxy
    then
    .all
  then
;

: DEL \ -- ; ^G del next
  bpos @  bstr +  < if ( del char)
    bpos @  dup 1+ swap  bufe bl over c!  over - 1+ cmove
    .rt  cxy  chgd
  else ( join line)
    chgd  leav  line  COLS over - >r
    +  dup 1+ zcount  r> min  rot zplace  .all
  then
;

: BS \ -- ; ^H del prev
  col @ if  lft  del  then
;

: DLN \ -- ; ^Y del line
  lin @  lines < if
    ladr @
    ld  enter  13 ladr -c!
    ladr !
    chgd  leav  enter
    lines-  lin 1-!
    delete-line
    ltop @  YBOT +  lines < if
      line@
      ltop @  YBOT +  setlin  0 YBOT goxy  .line
      line!
    then
    .pos
  then
;

: RST \ -- ; ^L restore line
  0 lpos @ goxy  clear-line  .line  cxy
  enter
;

: DEOL \ -- ; ^T delete to EOL
  bpos @ bufl blank
  clear-line chgd
;

: CHAROK \ c -- ; print char
  room? if
    dup emit
    ovr if
      bpos c!+
    else
      insc .rt
    then
    col 1+!   .pos   chgd
  end
  drop
;
  
: TAB \ -- ; ^I indent by spaces
  begin
    bl charok
  col @ TABM and 0= until
;

: CLR \ -- ; ^Z
  s" *** Clear text: " sure?
  if  leav  -txt  then
  .all
;

: /FID ( -- ) -1 to fid ;
/fid

: STNAM \ a u -- a u ;
  2dup fnam pack  count upper  /fid  dup 0= throw
;

: WAIT \ -- ;
  msg ." *** Wait"
;

: CLOSF \ -- ;
  fid  -1 of end  close-file drop  /fid
;

: (RD) \ a u -- ;
  stnam  r/w open-file if 2 throw then  to fid
  0 to lines  tbuf  dup off  dup 1+ nxt !
  begin ( a)
    dup COLS  2dup + mem u> throw
    fid read-line throw ( a u' f )
  while
    2dup >blanks  +  0 over c! ( null)
    1+  dup nxt !  lines+
  repeat  2drop  nxt @ off
;
  
: RD \ a u -- ;
  wait   leav 
  s" F" +ext  ( append .F extension if none )
  ['] (rd) catch ?dup if
    2 <> if s" load/size error" .err   0 fnam c! then
    2drop ctrc
  then
  closf  curtop  .all
;

: (WR) \ a u -- ;
  stnam  r/w create-file throw  to fid
  tbuf
  lines 0 ?do
    zcount  2dup fid write-line throw + 1+
  loop
  drop
;

: WR \ a u -- ;
  wait   leav
  ['] (wr) catch if
    2drop  s" save error" .err
  else
    twc off
  then
  closf .all
;

: SAV \ -- ;
  fnam count dup 0= if 2drop end  wr ;
  
: QUT \ -- ; ^KQ exit w/o save
  leav   xf on
  twc @ if
    s" *** Text changed: " sure? not
    if .hd   xf off then
  then
;

dup [if]
: HLP \ -- ; ^J help
  leav  page  14 spaces  ." Help Menu"
  cr  ." ^E   Up cursor      ^R   Prev page"
  cr  ." ^X   Down cursor    ^C   Next page"
  cr  ." ^D   Right cursor   ^G   Del next char"
  cr  ." ^S   Left cursor    ^H   Del prev char"
  cr  ." ^QS  Home cursor    ^QD  EOL cursor" 
  cr  ." ^QR  Begin cursor   ^QC  EOF cursor"
  cr  ." ^M   New line       ^T   Erase to EOL"
  cr  ." ^Y   Delete line    ^L   Restore line"
  cr  ." ^I   Indent         ^J   Help"
  cr  ." ^Z   Clear text     ^V   Ins/Ovr"
  cr  ." ^B   Redraw screen  ^N   Line number"
  cr  ." ^KS  Save"
  cr  ." ^KQ or Esc  Quit    ^KX  Exit with save"
  cr cr  akey  .all
;
[then]

: KMAP \ c1 -- c2 ; map in arrow keys etc
[ cpm? ] [if]
  $14E c@ of [ctrl] H end
  $14F c@ of [ctrl] E end
  $150 c@ of [ctrl] X end
  $151 c@ of [ctrl] D end
  $152 c@ of [ctrl] S end
  127     of [ctrl] G end
[else]
  char2 off
  $147 of [ctrl] Q [ctrl] S char2 ! end \ Home  = ^QS
  $148 of [ctrl] E end                  \ Up    = ^E
  $149 of [ctrl] R end                  \ PgUp  = ^R
  $14B of [ctrl] S end                  \ Left  = ^S
  $14D of [ctrl] D end                  \ Right = ^D
  $14F of [ctrl] Q [ctrl] D char2 ! end \ End   = ^QD
  $150 of [ctrl] X end                  \ Down  = ^X
  $151 of [ctrl] C end                  \ PgDn  = ^C
  $152 of [ctrl] V end                  \ Ins   = ^V
  $153 of [ctrl] G end                  \ Del   = ^G
  $175 of [ctrl] Q [ctrl] C char2 ! end \ ^End  = ^QC
  $177 of [ctrl] Q [ctrl] R char2 ! end \ ^Home = ^QR
[then]
;

: CTLQ \ -- ; ^Q
[ cpm? ]  [if]
  [char] Q ckey kmap
[else]
  char2 @ ?dup not if [char] Q ckey then
[then]
  [ctrl] S of hom end \ ^QS cursor home
  [ctrl] D of eol end \ ^QD cursor EOL
  [ctrl] R of bof end \ ^QR cursor begin of file
  [ctrl] C of eof end \ ^QC cursor EOF
  drop
;   

: CTLK ( -- ) \ ^K
  [char] K ckey
  [ctrl] Q of qut       end  \ ^KQ exit w/o save
  [ctrl] X of sav xf on end  \ ^KX exit with save
  [ctrl] S of sav       end  \ ^KS save
  drop
;   

: OVIN \ -- ; ^V
  ovr not to ovr  \ reverse ovewrite/insert flag
  .ovr \ print status
;
  
create KeyTable
' noop ,  ' noop , \  0  1 ; ^@  ^A
' .all ,  ' pd   , \  2  3 ; ^B  ^C
' rt   ,  ' upc  , \  4  5 ; ^D  ^E
' noop ,  ' del  , \  6  7 ; ^F  ^G
' bs   ,  ' tab  , \  8  9 ; ^H  ^I
' noop ,  ' ctlk , \ 10 11 ; ^J  ^K
' rst  ,  ' nln  , \ 12 13 ; ^L  ^M
' lnum ,  ' noop , \ 14 15 ; ^N  ^O
' noop ,  ' ctlq , \ 16 17 ; ^P  ^Q
' pu   ,  ' lft  , \ 18 19 ; ^R  ^S
' deol ,  ' noop , \ 20 21 ; ^T  ^U
' ovin ,  ' noop , \ 22 23 ; ^V  ^W
' dn   ,  ' dln  , \ 24 25 ; ^X  ^Y
' clr  ,  ' qut  , \ 26 27 ; ^Z  ^[ = esc
' noop ,  ' noop , \ 28 29 ; ^\  ^]
' noop ,  ' noop , \ 30 31 ; ^^  ^_

: CMD \ -- ;
  key kmap   maxm and
  dup bl < if
    cells KeyTable + @   execute
  else
    charok
  then
;

: (TED) \ col line caddr len -- ; load & edit textfile addr len
  setup  page  .hd  dup if \ if len=0 don't load
    rd  swap ( col ) XMAX min col !
    ( line ) 1- gln
  else
    2drop 2drop
  then
  enter  xf off
  begin cmd   xf @ until
  page
;

( *) [if]

\ Turnkey version
' hlp 10 cells KeyTable + ! \ ^J for Help

: TED \ -- ;
  0 1 cmdtail (ted)
;

turnkey ted ted bye

[else]

\ Resident version
-? : TED ( "filename[.F]" -- )
  token  dup if
    0 1  2swap
  else
    2drop  loadline 2@  lastfile
  then
  (ted)
  col @ lin @ 1+ loadline 2! \ return on leaving position at next start of TED
;

\ aka TED EDIT
behead cpm? cmd

[then]

forth definitions application
